from itertools import product
from typing import TYPE_CHECKING

import numpy as np

from scipy.optimize import linear_sum_assignment


if TYPE_CHECKING:
    from cogent3.core.tree import TreeNode


def get_tree_distance_measure(method: str, is_rooted: bool):
    if method in _TREE_DISTANCE_FUNCTIONS:
        return _TREE_DISTANCE_FUNCTIONS[method]

    if is_rooted and method in _ROOTED_TREE_DISTANCE_FUNCTIONS:
        return _ROOTED_TREE_DISTANCE_FUNCTIONS[method]

    if not is_rooted and method in _UNROOTED_TREE_DISTANCE_FUNCTIONS:
        return _UNROOTED_TREE_DISTANCE_FUNCTIONS[method]

    raise ValueError(
        f"Tree distance method '{method}' is not supported for {'' if is_rooted else 'un'}rooted trees."
    )


def unrooted_robinson_foulds(tree1: "TreeNode", tree2: "TreeNode") -> int:
    """Calculate the Robinson-Foulds distance between two unrooted trees.

    Parameters
    ----------
    tree1, tree2: TreeNode
        Trees to calculate the distance between.

    Returns
    -------
    int
        The unrooted Robinson-Foulds distance.

    Notes
    -----
    For unrooted trees, the Robinson-Foulds distance [1]_ is defined
    as the cardinality of the symmetric difference of the set of splits
    for the two trees.

    Trees should have matching tips and must not be rooted.

    References
    ----------
    .. [1] Robinson, David F., and Leslie R. Foulds.
       Comparison of phylogenetic trees.
       Mathematical biosciences 53.1-2 (1981): 131-147.
    """

    names = tree1.get_tip_names()

    if set(names) != set(tree2.get_tip_names()):
        raise ValueError("tree tip names must match")
    if len(tree1.children) == 2 or len(tree2.children) == 2:
        raise ValueError("trees must be unrooted")

    tree1_clusters = tree1.subsets()
    tree2_clusters = tree2.subsets()

    tree1_splits = _compute_splits(tree1_clusters, names)
    tree2_splits = _compute_splits(tree2_clusters, names)

    return len(tree1_splits.symmetric_difference(tree2_splits))


def lin_rajan_moret(tree1: "TreeNode", tree2: "TreeNode") -> int:
    """Calculate the Lin-Rajan-Moret distance (matching distance)
    between two unrooted trees.

    Parameters
    ----------
    tree1, tree2: TreeNode
        Trees to calculate the distance between.

    Returns
    -------
    int
        The Lin-Rajan-Moret distance.

    Notes
    -----
    The Lin-Rajan-Moret distance [1]_ displays superior statistical
    properties than the Robinson-Foulds distance [2]_
    on unrooted trees.

    Trees should have matching tips and must not be rooted.

    References
    ----------
    .. [1] Lin et al. 2012
       A Metric for Phylogenetic Trees Based on Matching
       IEEE/ACM Transactions on Computational Biology and Bioinformatics
       vol. 9, no. 4, pp. 1014-1022, July-Aug. 2012
    .. [2] Robinson, David F., and Leslie R. Foulds.
       Comparison of phylogenetic trees.
       Mathematical biosciences 53.1-2 (1981): 131-147.
    """
    names = tree1.get_tip_names()

    if set(names) != set(tree2.get_tip_names()):
        raise ValueError("tree tip names must match")
    if len(tree1.children) == 2 or len(tree2.children) == 2:
        raise ValueError("trees must be unrooted")

    names.sort()

    vector1 = _convert_tree_to_vectors(tree1, names)
    vector2 = _convert_tree_to_vectors(tree2, names)

    matching_distance = _matched_distance(vector1, vector2)

    return matching_distance


def rooted_robinson_foulds(tree1: "TreeNode", tree2: "TreeNode") -> int:
    """Calculate the Robinson-Foulds distance between two rooted trees.

    Parameters
    ----------
    tree1, tree2: TreeNode
        Trees to calculate the distance between.

    Returns
    -------
    int
        The rooted Robinson-Foulds distance

    Notes
    -----
    For rooted trees, the Robinson-Foulds distance [1]_ is defined as the
    cardinality of the symmetric difference of the set of clades for
    the two trees.

    Trees should have matching tips and must be rooted.

    References
    ----------
    .. [1] Robinson, David F., and Leslie R. Foulds.
       Comparison of phylogenetic trees.
       Mathematical biosciences 53.1-2 (1981): 131-147.
    """
    if set(tree1.get_tip_names()) != set(tree2.get_tip_names()):
        raise ValueError("tree tip names must match")
    if len(tree1.children) != 2 or len(tree2.children) != 2:
        raise ValueError("trees must be rooted")

    tree1_clusters = tree1.subsets()
    tree2_clusters = tree2.subsets()

    return len(tree1_clusters.symmetric_difference(tree2_clusters))


def matching_cluster_distance(tree1: "TreeNode", tree2: "TreeNode") -> int:
    """Calculate the Matching Cluster distance between two rooted trees.

    Parameters
    ----------
    tree1, tree2: TreeNode
        Trees to calculate the distance between.

    Returns
    -------
    int
        The Matching Cluster distance.

    Notes
    -----
    The Matching Cluster distance [1]_ is a similar metric
    to the beta-distance [2]_.

    The Matching Cluster distance [1]_ displays superior statistical
    properties than the Robinson-Foulds distance [3]_ on rooted trees.

    Trees should have matching tips and must be rooted.

    References
    ----------
    .. [1] Bogdanowicz, D., & Giaro, K. (2013).
       On a matching distance between rooted phylogenetic trees.
       International Journal of Applied Mathematics and Computer Science, 23(3), 669-684.
    .. [2] Boorman, S. A., & Olivier, D. C. (1973).
       Metrics on spaces of finite trees.
       Journal of Mathematical Psychology, 10(1), 26-59.
    .. [3] Robinson, David F., and Leslie R. Foulds.
       Comparison of phylogenetic trees.
       Mathematical biosciences 53.1-2 (1981): 131-147.
    """

    if set(tree1.get_tip_names()) != set(tree2.get_tip_names()):
        raise ValueError("tree tip names must match")
    if len(tree1.children) != 2 or len(tree2.children) != 2:
        raise ValueError("trees must be rooted")

    tree1_clusters = list(tree1.subsets())
    tree2_clusters = list(tree2.subsets())

    while len(tree1_clusters) < len(tree2_clusters):
        tree1_clusters.append(set())
    while len(tree2_clusters) < len(tree1_clusters):
        tree2_clusters.append(set())

    adjacency = np.zeros(shape=(len(tree1_clusters), len(tree2_clusters)))

    for i, cluster_1 in enumerate(tree1_clusters):
        for j, cluster_2 in enumerate(tree2_clusters):
            adjacency[i, j] = len(cluster_1.symmetric_difference(cluster_2))

    row_ind, col_ind = linear_sum_assignment(adjacency)
    distance = int(adjacency[row_ind, col_ind].sum())

    return distance


def _convert_tree_to_vectors(tree: "TreeNode", tip_names: list) -> np.ndarray:
    ref_tip = tip_names[0]
    name_set = set(tip_names)
    name_index = {n: i for i, n in enumerate(tip_names)}
    # we get the tree as a set of splits.
    splits = tree.subsets()
    rows = np.zeros((len(splits), len(tip_names)), dtype=bool)
    for i, split in enumerate(splits):
        row = rows[i]
        # Cogent only returns one side of a
        # split, so we build the other side
        if ref_tip in split:
            names = list(split)
        else:
            names = list(name_set - split)
        indices = [name_index[n] for n in names]
        row[indices] = True
    return rows


def _hamming(vector1: np.ndarray, vector2: np.ndarray) -> int:
    return (vector1 != vector2).sum()


def _weight(vector1: np.ndarray, vector2: np.ndarray) -> int:
    return min(_hamming(vector1, vector2), _hamming(vector1, 1 - vector2))


def _bipartite_graph(vector1: np.ndarray, vector2: np.ndarray) -> np.ndarray:
    if not len(vector1) == len(vector2):
        raise ValueError("number of edges must be equal")

    B = np.empty([len(vector1)] * 2, int)
    for i, j in product(*[range(len(vector1))] * 2):
        B[i, j] = _weight(vector1[i], vector2[j])
    return B


def _matched_distance(vector1: np.ndarray, vector2: np.ndarray) -> int:
    B = _bipartite_graph(vector1, vector2)
    matching = linear_sum_assignment(B)
    return B[matching].sum()


def _compute_splits(clades: frozenset[frozenset], tip_names: list) -> set[frozenset]:
    ref_tip = tip_names[0]
    name_set = frozenset(tip_names)

    splits = set()
    for clade in clades:
        if ref_tip in clade:
            splits.add(clade)
        else:
            splits.add(name_set - clade)
    return splits


_TREE_DISTANCE_FUNCTIONS = {
    "rooted_robinson_foulds": rooted_robinson_foulds,
    "unrooted_robinson_foulds": unrooted_robinson_foulds,
    "matching_cluster": matching_cluster_distance,
    "lin_rajan_moret": lin_rajan_moret,
    "rrf": rooted_robinson_foulds,
    "urf": unrooted_robinson_foulds,
    "mc": matching_cluster_distance,
    "lrm": lin_rajan_moret,
}

_ROOTED_TREE_DISTANCE_FUNCTIONS = {
    "rf": rooted_robinson_foulds,
    "matching": matching_cluster_distance,
}

_UNROOTED_TREE_DISTANCE_FUNCTIONS = {
    "rf": unrooted_robinson_foulds,
    "matching": lin_rajan_moret,
}



#!/usr/bin/env python
""" Unit tests for Genetic Code classes.
"""
from unittest import TestCase

import pytest

from cogent3 import DNA, RNA
from cogent3.core.genetic_code import (
    DEFAULT,
    GeneticCode,
    GeneticCodeInitError,
    GeneticCodes,
    InvalidCodonError,
    available_codes,
    get_code,
)


class GeneticCodeTests(TestCase):
    """Tests of the GeneticCode class."""

    def setUp(self):
        """Set up some standard genetic code representations."""
        self.SGC = "FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG"
        self.mt = "FFLLSSSSYY**CCWWLLLLPPPPHHQQRRRRIIMMTTTTNNKKSS**VVVVAAAADDEEGGGG"
        self.AllG = "GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG"

        self.WrongLength = [
            "GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG" "",
            "GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG",
        ]
        self.NcbiStandard = [
            "FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
            1,
            "Standard",
            "---M---------------M---------------M----------------------------",
        ]

    def test_init(self):
        """GeneticCode init should work with correct-length sequences"""
        sgc = GeneticCode(self.SGC)
        self.assertEqual(sgc["UUU"], "F")
        mt = GeneticCode(self.mt)
        self.assertEqual(mt["UUU"], "F")
        allg = GeneticCode(self.AllG)
        self.assertEqual(allg["UUU"], "G")
        for i in self.WrongLength:
            self.assertRaises(GeneticCodeInitError, GeneticCode, i)

    def test_standard_code(self):
        """Standard genetic code from NCBI should have correct properties"""
        sgc = GeneticCode(*self.NcbiStandard)
        self.assertEqual(
            sgc.code_sequence,
            "FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
        )
        self.assertEqual(
            sgc.start_codon_sequence,
            "---M---------------M---------------M----------------------------",
        )
        self.assertEqual(sgc.start_codons, {"TTG": "M", "CTG": "M", "ATG": "M"})
        self.assertEqual(sgc.ID, 1)
        self.assertEqual(sgc.name, "Standard")
        self.assertEqual(sgc["UUU"], "F")
        self.assertEqual(sgc.is_start("ATG"), True)
        self.assertEqual(sgc.is_start("AAA"), False)
        self.assertEqual(sgc.is_stop("UAA"), True)
        self.assertEqual(sgc.is_stop("AAA"), False)
        self.assertEqual(len(sgc.sense_codons), 61)
        self.assertIn("AAA", sgc.sense_codons)
        self.assertNotIn("TGA", sgc.sense_codons)

    def test_standard_code_lookup(self):
        """GeneticCodes should hold codes keyed by id as string and number"""
        sgc_new = GeneticCode(*self.NcbiStandard)
        sgc_number = GeneticCodes[1]
        sgc_string = GeneticCodes["1"]
        for sgc in sgc_new, sgc_number, sgc_string:
            self.assertEqual(
                sgc.code_sequence,
                "FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
            )
            self.assertEqual(
                sgc.start_codon_sequence,
                "---M---------------M---------------M----------------------------",
            )
            self.assertEqual(sgc.start_codons, {"TTG": "M", "CTG": "M", "ATG": "M"})
            self.assertEqual(sgc.ID, 1)
            self.assertEqual(sgc.name, "Standard")
            self.assertEqual(sgc["TTT"], "F")
            self.assertEqual(sgc.is_start("ATG"), True)
            self.assertEqual(sgc.is_start("AAA"), False)
            self.assertEqual(sgc.is_stop("TAA"), True)
            self.assertEqual(sgc.is_stop("AAA"), False)

        mtgc = GeneticCodes[2]
        self.assertEqual(mtgc.name, "Vertebrate Mitochondrial")
        self.assertEqual(mtgc.is_start("AUU"), True)
        self.assertEqual(mtgc.is_stop("UGA"), False)

        self.assertEqual(
            sgc_new.changes(mtgc), {"AGA": "R*", "AGG": "R*", "ATA": "IM", "TGA": "*W"}
        )
        self.assertEqual(
            mtgc.changes(sgc_new), {"AGA": "*R", "AGG": "*R", "ATA": "MI", "TGA": "W*"}
        )
        self.assertEqual(mtgc.changes(mtgc), {})
        self.assertEqual(
            mtgc.changes(
                "FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG"
            ),
            {"AGA": "*R", "AGG": "*R", "ATA": "MI", "TGA": "W*"},
        )

    def test_str(self):
        """GeneticCode str() should return its code string"""
        code_strings = self.SGC, self.mt, self.AllG
        codes = list(map(GeneticCode, code_strings))
        for code, string in zip(codes, code_strings):
            self.assertEqual(str(code), string)
        # check an example directly in case strings are bad
        self.assertEqual(
            str(self.SGC),
            "FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG",
        )

    def test_cmp(self):
        """GeneticCode cmp() should act on code strings"""
        sgc_1 = GeneticCode(self.SGC)
        sgc_2 = GeneticCode(self.SGC)
        self.assertEqual(sgc_1 is sgc_2, False)  # ensure different objects
        # self.assertNotEqual(sgc_1, sgc_2) # GREG
        self.assertEqual(sgc_1, sgc_2)
        mtgc = GeneticCode(self.mt)
        self.assertNotEqual(sgc_1, mtgc)

    def test_getitem_codon(self):
        """GeneticCode getitem should return amino acid for codon"""
        # specific checks of a particular codon in the standard code
        variant_codons = ["AUU", "AUU", "AUU", "ATT", "ATU", "ATU"]
        sgc = GeneticCode(self.SGC)
        for i in variant_codons:
            self.assertEqual(sgc[i], "I")
        # full check for the standard code
        codons = [a + b + c for a in "UCAG" for b in "TCAG" for c in "UCAG"]
        for codon, aa in zip(codons, self.SGC):
            self.assertEqual(sgc[codon], aa)
        # full check for another code
        allg = GeneticCode(self.AllG)
        for codon, aa in zip(codons, self.AllG):
            self.assertEqual(allg[codon], aa)
        # check that degenerate codon returns X
        self.assertEqual(sgc["NNN"], "X")

    def test_getitem_aa(self):
        """GeneticCode getitem should return codon set for aa"""
        # for all G, should return all the codons (in some order)
        allg = GeneticCode(self.AllG)
        codons = [a + b + c for a in "TCAG" for b in "TCAG" for c in "TCAG"]
        g_codons = allg["G"]
        codons_copy = codons[:]
        self.assertEqual(g_codons, codons_copy)

        # check some known cases in the standard genetic code
        sgc = GeneticCode(self.SGC)
        exp_ile = ["ATT", "ATC", "ATA"]
        obs_ile = sgc["I"]
        self.assertEqual(obs_ile, exp_ile)

        sgc["R"]
        self.assertEqual(obs_ile, exp_ile)

        exp_leu = ["TTA", "TTG", "CTT", "CTC", "CTA", "CTG"]
        obs_leu = sgc["L"]
        self.assertEqual(obs_leu, exp_leu)

        exp_met = ["ATG"]
        obs_met = sgc["M"]
        self.assertEqual(obs_met, exp_met)

        # unknown aa should return []
        self.assertEqual(sgc["U"], [])

    def test_getitem_invalid_length(self):
        """GeneticCode getitem should raise InvalidCodonError on wrong length"""
        sgc = GeneticCode(self.SGC)
        self.assertRaises(InvalidCodonError, sgc.__getitem__, "AAAA")
        self.assertRaises(InvalidCodonError, sgc.__getitem__, "AA")

    def test_Blocks(self):
        """GeneticCode Blocks should return correct list"""
        sgc = GeneticCode(self.SGC)
        exp_blocks = [
            ["TTT", "TTC"],
            ["TTA", "TTG"],
            ["TCT", "TCC", "TCA", "TCG"],
            ["TAT", "TAC"],
            ["TAA", "TAG"],
            ["TGT", "TGC"],
            ["TGA"],
            ["TGG"],
            ["CTT", "CTC", "CTA", "CTG"],
            ["CCT", "CCC", "CCA", "CCG"],
            ["CAT", "CAC"],
            ["CAA", "CAG"],
            ["CGT", "CGC", "CGA", "CGG"],
            ["ATT", "ATC"],
            ["ATA"],
            ["ATG"],
            ["ACT", "ACC", "ACA", "ACG"],
            ["AAT", "AAC"],
            ["AAA", "AAG"],
            ["AGT", "AGC"],
            ["AGA", "AGG"],
            ["GTT", "GTC", "GTA", "GTG"],
            ["GCT", "GCC", "GCA", "GCG"],
            ["GAT", "GAC"],
            ["GAA", "GAG"],
            ["GGT", "GGC", "GGA", "GGG"],
        ]
        self.assertEqual(sgc.blocks, exp_blocks)

    def test_Anticodons(self):
        """GeneticCode anticodons should return correct list"""
        sgc = GeneticCode(self.SGC)
        exp_anticodons = {
            "F": ["AAA", "GAA"],
            "L": ["TAA", "CAA", "AAG", "GAG", "TAG", "CAG"],
            "Y": ["ATA", "GTA"],
            "*": ["TTA", "CTA", "TCA"],
            "C": ["ACA", "GCA"],
            "W": ["CCA"],
            "S": ["AGA", "GGA", "TGA", "CGA", "ACT", "GCT"],
            "P": ["AGG", "GGG", "TGG", "CGG"],
            "H": ["ATG", "GTG"],
            "Q": ["TTG", "CTG"],
            "R": ["ACG", "GCG", "TCG", "CCG", "TCT", "CCT"],
            "I": ["AAT", "GAT", "TAT"],
            "M": ["CAT"],
            "T": ["AGT", "GGT", "TGT", "CGT"],
            "N": ["ATT", "GTT"],
            "K": ["TTT", "CTT"],
            "V": ["AAC", "GAC", "TAC", "CAC"],
            "A": ["AGC", "GGC", "TGC", "CGC"],
            "D": ["ATC", "GTC"],
            "E": ["TTC", "CTC"],
            "G": ["ACC", "GCC", "TCC", "CCC"],
        }
        self.assertEqual(sgc.anticodons, exp_anticodons)

    def test_translate(self):
        """GeneticCode translate should return correct amino acid string"""
        allg = GeneticCode(self.AllG)
        sgc = GeneticCode(self.SGC)
        mt = GeneticCode(self.mt)

        seq = "AUGCAUGACUUUUGA"
        #      .  .  .  .  .        markers for codon start
        self.assertEqual(allg.translate(seq), "GGGGG")
        self.assertEqual(allg.translate(seq, 1), "GGGG")
        self.assertEqual(allg.translate(seq, 2), "GGGG")
        self.assertEqual(allg.translate(seq, 3), "GGGG")
        self.assertEqual(allg.translate(seq, 4), "GGG")
        self.assertEqual(allg.translate(seq, 12), "G")
        self.assertEqual(allg.translate(seq, 14), "")
        self.assertRaises(ValueError, allg.translate, seq, 15)
        self.assertRaises(ValueError, allg.translate, seq, 20)

        self.assertEqual(sgc.translate(seq), "MHDF*")
        self.assertEqual(sgc.translate(seq, 3), "HDF*")
        self.assertEqual(sgc.translate(seq, 6), "DF*")
        self.assertEqual(sgc.translate(seq, 9), "F*")
        self.assertEqual(sgc.translate(seq, 12), "*")
        self.assertEqual(sgc.translate(seq, 14), "")
        # check shortest translatable sequences
        self.assertEqual(sgc.translate("AAA"), "K")
        self.assertEqual(sgc.translate(""), "")

        # check that different code gives different results
        self.assertEqual(mt.translate(seq), "MHDFW")

        # check translation with invalid codon(s)
        self.assertEqual(sgc.translate("AAANNNCNC123UUU"), "KXXXF")

    def test_sixframes(self):
        """GeneticCode sixframes should provide six-frame translation"""

        class fake_rna(str):
            """Fake RNA class with reverse-complement"""

            def __new__(cls, seq, rev):
                return str.__new__(cls, seq)

            def __init__(self, seq, rev):
                self.seq = seq
                self.rev = rev

            def rc(self):
                return self.rev

        test_rna = fake_rna("AUGCUAACAUAAA", "UUUAUGUUAGCAU")
        #                    .  .  .  .  .    .  .  .  .  .
        sgc = GeneticCode(self.SGC)
        self.assertEqual(
            sgc.sixframes(test_rna), ["MLT*", "C*HK", "ANI", "FMLA", "LC*H", "YVS"]
        )

        # should also actually work with an RNA or DNA sequence!!!
        test_rna = RNA.make_seq("AUGCUAACAUAAA")
        self.assertEqual(
            sgc.sixframes(test_rna), ["MLT*", "C*HK", "ANI", "FMLA", "LC*H", "YVS"]
        )

    def test_stop_indexes(self):
        """should return stop codon indexes for a specified frame"""
        sgc = GeneticCode(self.SGC)
        seq = DNA.make_seq("ATGCTAACATAAA")
        expected = [[9], [4], []]
        for frame, expect in enumerate(expected):
            got = sgc.get_stop_indices(seq, start=frame)
            self.assertEqual(got, expect)

    def test_Synonyms(self):
        """GeneticCode synonyms should return aa -> codon set mapping."""
        expected_synonyms = {
            "A": ["GCT", "GCC", "GCA", "GCG"],
            "C": ["TGT", "TGC"],
            "D": ["GAT", "GAC"],
            "E": ["GAA", "GAG"],
            "F": ["TTT", "TTC"],
            "G": ["GGT", "GGC", "GGA", "GGG"],
            "H": ["CAT", "CAC"],
            "I": ["ATT", "ATC", "ATA"],
            "K": ["AAA", "AAG"],
            "L": ["TTA", "TTG", "CTT", "CTC", "CTA", "CTG"],
            "M": ["ATG"],
            "N": ["AAT", "AAC"],
            "P": ["CCT", "CCC", "CCA", "CCG"],
            "Q": ["CAA", "CAG"],
            "R": ["AGA", "AGG", "CGT", "CGC", "CGA", "CGG"],
            "S": ["TCT", "TCC", "TCA", "TCG", "AGT", "AGC"],
            "T": ["ACT", "ACC", "ACA", "ACG"],
            "V": ["GTT", "GTC", "GTA", "GTG"],
            "W": ["TGG"],
            "Y": ["TAT", "TAC"],
            "*": ["TAA", "TAG", "TGA"],
        }
        obs_synonyms = GeneticCode(self.SGC).synonyms
        # note that the lists will be arbitrary-order
        for i in expected_synonyms:
            self.assertCountEqual(obs_synonyms[i], expected_synonyms[i])

    def test_get_code(self):
        """correctly return the genetic code"""
        for code_id in [1, "1", "Standard", DEFAULT]:
            got = get_code(code_id)
            self.assertEqual(got, DEFAULT)
        got = get_code(2)
        self.assertEqual(got.name.lower(), "vertebrate mitochondrial")

    def test_available_codes(self):
        """avaialable_codes returns Table with correct shape"""
        codes = available_codes()
        self.assertTrue(codes.shape[0] > 17)
        self.assertEqual(codes[0, "Name"], "Standard")

    def test_to_table(self):
        """tests to_table method."""
        from cogent3.core.moltype import IUPAC_PROTEIN_code_aa

        sgc = GeneticCode(self.SGC)
        table = sgc.to_table()
        # check num rows
        self.assertEqual(table.shape[0], len(IUPAC_PROTEIN_code_aa))
        # check there are 3 headers
        self.assertEqual(table.shape[1], 3)

    def test_to_regex(self):
        """creates a regex from aa seq to match a DNA sequence"""
        import re

        from cogent3 import make_seq

        dna = "ACCGAACAGGGC"
        aa = "TEQG"
        pattern = DEFAULT.to_regex(aa)
        self.assertTrue("".join(re.findall(pattern, dna)) == dna)
        # note that Z is Q or E
        aa = "TZQG"
        pattern = DEFAULT.to_regex(aa)
        self.assertTrue("".join(re.findall(pattern, dna)) == dna)
        aa = make_seq(aa, moltype="protein")
        pattern = DEFAULT.to_regex(aa)
        self.assertTrue("".join(re.findall(pattern, dna)) == dna)

    def test_repr_html(self):
        """exercising the _repr_html_ method"""
        gc = get_code(1)
        got = gc._repr_html_().strip()
        self.assertTrue(
            '<div class="c3table">' in got or '<div class="c3align">' in got
        )
        self.assertTrue("<table>" in got)
        self.assertTrue("</table>" in got)
        self.assertIn("Standard", got)


# starting pytest versions
@pytest.mark.parametrize("code", (1, 2))
def test_get_alphabet(code):
    from cogent3.core.alphabet import Alphabet

    gc = get_code()
    alpha_w_stop = gc.get_alphabet(include_stop=True)
    assert tuple(alpha_w_stop) == tuple(gc.codons)
    assert isinstance(alpha_w_stop, Alphabet)
    alpha_wo_stop = gc.get_alphabet(include_stop=False)
    assert tuple(alpha_wo_stop) == tuple(gc.sense_codons)
    assert isinstance(alpha_wo_stop, Alphabet)



#!/usr/bin/env python

from .align import (
    _align_pairwise,
    classic_align_pairwise,
    global_pairwise,
    local_pairwise,
    make_dna_scoring_dict,
    make_generic_scoring_dict,
)


__all__ = [
    "align",
    "dp_calculation",
    "indel_model",
    "indel_positions",
    "pairwise",
    "progressive",
    "pycompare",
    "traceback",
]



from numpy import exp, log

from cogent3.util.io import atomic_write

from . import consensus


class _UserList(list):
    def __getitem__(self, index):
        # Helpful to keep type after truncation like [self[:10]],
        # but not self[0] or self[x,y,-1]
        result = list.__getitem__(self, index)
        if isinstance(index, slice) and index.step is None:
            result = type(self)(result)
        return result


class ScoredTreeCollection(_UserList):
    """An ordered list of (score, tree) tuples"""

    def write(self, filename):
        with atomic_write(filename, mode="wt") as f:
            for score, tree in self:
                f.writelines(
                    self.scored_tree_format(
                        tree.get_newick(with_distances=True), str(score)
                    )
                )

    def scored_tree_format(self, tree, score):
        return [tree, "\t[", score, "]\n"]

    def get_consensus_tree(self, strict=None, method="unrooted"):
        ctrees = self.get_consensus_trees(strict, method=method)
        assert len(ctrees) == 1, len(ctrees)
        return ctrees[0]

    def get_consensus_trees(self, strict=True, method="unrooted"):
        if strict is None:
            strict = True
        return consensus.weighted_majority_rule(self, strict, method=method)


class UsefullyScoredTreeCollection(ScoredTreeCollection):
    def scored_tree_format(self, tree, score):
        return [score, "\t", tree, "\n"]


class WeightedTreeCollection(UsefullyScoredTreeCollection):
    """An ordered list of (weight, tree) tuples"""

    def get_consensus_trees(self, strict=False, method="unrooted"):
        if strict is None:
            strict = False
        return consensus.weighted_majority_rule(self, strict, method=method)


class LogLikelihoodScoredTreeCollection(UsefullyScoredTreeCollection):
    """An ordered list of (log likelihood, tree) tuples"""

    def __init__(self, trees):
        list.__init__(self, trees)
        # Quick and very dirty check of order
        assert self[0][0] >= self[-1][0]

    def get_consensus_tree(self, cutoff=None, strict=False, alpha=0.05):
        """See documentation for get_consensus_trees"""
        return self.get_consensus_trees(cutoff, strict, alpha)[0]

    def get_consensus_trees(self, cutoff=None, strict=False, alpha=0.05):
        """Returns a weighted consensus tree as described in Holland (2006).
        Weights transformed according to the class IV transformation in Jermiin
        (1997).

        Args:
            cutoff: Discard trees with weights that sum to cutoff.
            strict: Discard splits with consensus weight <= 0.5.
            alpha: Significance level, see Jermiiin (1997).

        Returns:
            Single consensus tree, in a list for backward compatibility.

        Holland, B. R., Jermiin, L. S., Moulton, V., and
        Investigators, S. T.-N. Y. (2006). Proceedings of the SMBE
        Tri-National Young Investigators' Workshop 2005. Improved consensus
        network techniques for genome-scale phylogeny. Mol Biol Evol,
        23(5), 848-855. doi:10.1093/molbev/msj061

        Jermiin, L. S., Olsen, G. J., Mengerson, K., and Easteal, S. (1997).
        Majority-rule consensus of phylogenetic trees obtained by
        maximum-likelihood analysis. Molecular Biology and Evolution,
        14(12):1296.
        """
        return self.get_weighted_trees(cutoff, alpha).get_consensus_trees(strict)

    def get_weighted_trees(self, cutoff=None, alpha=0.05):
        if cutoff is None:
            cutoff = 0.99
        assert 0 <= cutoff <= 1.0
        max_lnL = self[0][0]
        forgotten = log(alpha) / (self[-1][0] - max_lnL)
        weights = [exp(forgotten * (lnL - max_lnL)) for (lnL, t) in self]
        # add from smallest end to avoid rounding errors
        weights.reverse()
        tail = (1.0 - cutoff) * sum(weights)
        dropped = 0.0
        for index, weight in enumerate(weights):
            dropped += weight
            if dropped > tail:
                weights = weights[index:]
                break
        denominator = sum(weights)
        weights.reverse()
        return WeightedTreeCollection(
            (weight / denominator, tree) for (weight, (lnL, tree)) in zip(weights, self)
        )


def make_trees(filename):
    """Parse a file of (score, tree) lines. Scores can be positive probabilities
    or negative log likelihoods."""
    from cogent3 import make_tree

    infile = open(filename, "r")
    trees = []
    klass = list
    # expect score, tree
    for line in infile:
        line = line.split(None, 1)
        lnL = float(line[0])
        if lnL > 1:
            raise ValueError(f"likelihoods expected, not {lnL}")
        elif lnL > 0:
            assert klass in [list, WeightedTreeCollection]
            klass = WeightedTreeCollection
        else:
            assert klass in [list, LogLikelihoodScoredTreeCollection]
            klass = LogLikelihoodScoredTreeCollection
        tree = make_tree(treestring=line[1])
        trees.append((lnL, tree))
    trees.sort(reverse=True)
    infile.close()
    return klass(trees)



from __future__ import annotations

import contextlib
import inspect
import re
import textwrap
import warnings

import stevedore

from cogent3.util.table import Table

from .composable import is_app, is_app_composable
from .io import open_data_store


# Entry_point for apps to register themselves as plugins
APP_ENTRY_POINT = "cogent3.app"


def _get_extension_attr(extension):
    """
    This function returns app details for display.

    Notes
    -----
    This function also loads the module the app is in.
    """

    obj = extension.plugin

    if not is_app(obj):
        warnings.warn(
            f"{obj!r} from {obj.__module__!r} is not a valid cogent3 app, skipping"
        )

    _types = _make_types(obj)

    return [
        extension.module_name,
        extension.name,
        is_app_composable(obj),
        _doc_summary(obj.__doc__ or ""),
        ", ".join(sorted(_types["_data_types"])),
        ", ".join(sorted(_types["_return_types"])),
    ]


def _make_types(app) -> dict:
    """returns type hints for the input and output"""
    _types = {"_data_types": [], "_return_types": []}
    for tys in _types:
        types = getattr(app, tys, None) or []
        types = [types] if type(types) == str else types
        _types[tys] = [{None: ""}.get(e, e) for e in types]
    return _types


# private global to hold an ExtensionManager instance
__apps = None


def get_app_manager() -> stevedore.ExtensionManager:
    """
    Lazy load a stevedore ExtensionManager to collect apps.
    """
    global __apps
    if not __apps:
        __apps = stevedore.ExtensionManager(
            namespace=APP_ENTRY_POINT, invoke_on_load=False
        )

    return __apps


def available_apps(name_filter: str | None = None) -> Table:
    """
    returns Table listing the available apps
    """
    from cogent3.util.table import Table

    rows = []

    extensions = get_app_manager()

    for extension in extensions:
        if name_filter and name_filter not in extension.name:
            continue

        with contextlib.suppress(AttributeError):
            # probably a local scope issue in testing!
            rows.append(_get_extension_attr(extension))

    header = ["module", "name", "composable", "doc", "input type", "output type"]
    return Table(header=header, data=rows)


_get_param = re.compile('(?<=").+(?=")')


def _make_signature(app: type) -> str:
    from cogent3.util.misc import get_object_provenance

    if app is None:
        raise ValueError("app cannot be None")

    # if app is an instance, get the underlying class
    if not inspect.isclass(app):
        app = app.__class__

    init_sig = inspect.signature(app.__init__)
    app_name = app.__name__
    params = [f'"{app_name}"']
    empty_default = inspect._empty
    for k, v in init_sig.parameters.items():
        if k == "self":
            continue

        txt = repr(v).replace("\n", " ")
        # clean up text when callable() used  as a type hint
        txt = txt.replace("<built-in function callable>", "callable")

        val = _get_param.findall(txt)[0]
        if v.default is not empty_default and callable(v.default):
            val = val.split("=", maxsplit=1)
            if hasattr(v.default, "app_type"):
                val[-1] = f" {str(v.default)}"
            else:
                val[-1] = f" {get_object_provenance(v.default)}"
            val = "=".join(val)
        params.append(val.replace("\n", " "))

    sig_prefix = f"{app_name}_app = get_app"
    sig_prefix_length = len(sig_prefix)

    if len(", ".join(params)) + sig_prefix_length <= 68:  # plus 2 parentheses makes 70
        params = ", ".join(params)
        return f"{sig_prefix}({params})"

    indent = " " * 4
    params = ",\n".join([f"{indent}{p}" for p in params])

    return f"{sig_prefix}(\n{params},\n)"


def _doc_summary(doc):
    """return first para of docstring"""
    result = []
    for line in doc.splitlines():
        if line := line.strip():
            result.append(line)
        else:
            break
    return " ".join(result)


def _get_app_matching_name(name: str):
    """name can include module name"""

    if "." in name:
        modname, name = name.rsplit(".", maxsplit=1)
    else:
        modname = None

    extensions_matching = [
        extension for extension in get_app_manager() if extension.name == name
    ]
    if not extensions_matching:
        raise ValueError(f"App {name!r} not found. Please check for typos.")

    if modname:
        for extension in extensions_matching:
            if extension.module_name.endswith(modname):
                return extension.plugin
        raise ValueError(f"App {name!r} not found. Please check for typos.")

    elif len(extensions_matching) == 1:
        return extensions_matching[0].plugin
    else:
        raise NameError(
            f"Too many apps matching name {name!r},\n{available_apps().filtered(lambda x: name == x, columns='name')}"
        )


def get_app(_app_name: str, *args, **kwargs):
    """returns app instance, use app_help() to display arguments

    Raises
    ------
    NameError when multiple apps have the same name. In that case use a
    qualified class name.
    """
    return _get_app_matching_name(_app_name)(*args, **kwargs)


def _make_head(text: str) -> list[str]:
    """makes a restructured text formatted header"""
    return [text, "-" * len(text)]


def _clean_params_docs(text: str) -> str:
    """remove unnecessary indentation"""
    text = text.splitlines(keepends=False)
    prefix = re.compile(r"^\s{8}")  # expected indentation of constructor doc
    doc = []
    for line in text:
        line = prefix.sub("", line)
        doc.append(line)

    # Remove empty lines at the beginning and end of docstrings
    if not doc[0]:
        doc.pop(0)
    if not doc[-1]:
        doc.pop()

    return "\n".join(doc)


def _clean_overview(text: str) -> str:
    text = text.split()
    return "\n".join(textwrap.wrap(" ".join(text), break_long_words=False))


def _make_apphelp_docstring(app):
    docs = []
    app_doc = app.__doc__ or ""
    if app_doc.strip():
        docs.extend(_make_head("Overview") + [_clean_overview(app_doc)])

    docs.extend(_make_head("Options for making the app") + [_make_signature(app)])

    init_doc = app.__init__.__doc__ or ""
    if init_doc.strip():
        docs.extend(["", _clean_params_docs(init_doc)])

    types = _make_types(app)
    docs.extend([""] + _make_head("Input type") + [", ".join(types["_data_types"])])
    docs.extend([""] + _make_head("Output type") + [", ".join(types["_return_types"])])

    return "\n".join(docs)


def app_help(name: str):
    """displays help for the named app

    Parameters
    ----------
    name
        app name, e.g. 'min_length', or can include module information,
        e.g. 'cogent3.app.sample.min_length' or 'sample.min_length'. Use the
        latter (qualified class name) style when there are multiple matches
        to name.
    """
    app = _get_app_matching_name(name)
    print(_make_apphelp_docstring(app))



#!/usr/bin/env python
"""Writer for FASTA sequence format
"""

from typing import Optional

from cogent3.format.util import _AlignmentFormatter


def alignment_to_fasta(
    alignment_dict: dict[str, str],
    block_size: int = 60,
    order: Optional[list[str]] = None,
) -> str:
    """Returns a Fasta string given an alignment.

    Parameters
        ----------
        alignment_dict
            dict of seq_name
        block_size
            the sequence length to write to each line,
            by default 60
        order
            optional list of sequence names, which order to print in.
            Assumes complete and correct list of names,
            by default None

        Returns
        -------
        The alignment in the Fasta format.
    """
    order = order or []
    return FastaFormatter().format(alignment_dict, block_size, order)


class FastaFormatter(_AlignmentFormatter):
    def format(
        self,
        alignment_dict: dict[str, str],
        block_size: int = 60,
        order: Optional[list[str]] = None,
    ) -> str:
        """Format the alignment to Fasta.

        Parameters
        ----------
        alignment_dict
            dict of seq_name
        block_size
            the sequence length to write to each line,
            by default 60
        order
            optional list of sequence names, which order to print in.
            Assumes complete and correct list of names,
            by default None

        Returns
        -------
        The alignment in the Fasta format.
        """
        # setup
        if not order:
            order = list(alignment_dict.keys())
        self.set_align_info(alignment_dict, order)
        self.set_block_size(block_size)

        if len(alignment_dict) == 0:
            return ""

        return "".join(
            [
                ">%s\n%s"
                % (
                    seq,
                    self.wrap_string_to_block_size(
                        alignment_dict[seq], alt_block_size=block_size
                    ),
                )
                for seq in self.align_order
            ]
        )



"""Unit tests for distance_transform.py functions.
"""

from random import randint, sample, seed
from unittest import TestCase

from numpy import array, ones, shape, sqrt
from numpy.testing import assert_allclose

from cogent3.maths.distance_transform import (
    binary_dist_chisq,
    binary_dist_chord,
    binary_dist_euclidean,
    binary_dist_hamming,
    binary_dist_lennon,
    binary_dist_ochiai,
    binary_dist_otu_gain,
    binary_dist_pearson,
    binary_dist_sorensen_dice,
    dist_abund_jaccard,
    dist_bray_curtis,
    dist_bray_curtis_faith,
    dist_bray_curtis_magurran,
    dist_canberra,
    dist_chisq,
    dist_chord,
    dist_euclidean,
    dist_gower,
    dist_hellinger,
    dist_kulczynski,
    dist_manhattan,
    dist_morisita_horn,
    dist_pearson,
    dist_soergel,
    dist_spearman_approx,
    dist_specprof,
    jaccard,
    trans_chisq,
    trans_chord,
    trans_hellinger,
    trans_specprof,
    zeros,
)


class functionTests(TestCase):
    """Tests of top-level functions."""

    def setUp(self):
        self.mat_test = array([[10, 10, 20], [10, 15, 10], [15, 5, 5]], "float")

        self.emptyarray = array([], "d")
        self.mtx1 = array([[1, 3], [0.0, 23.1]], "d")
        self.dense1 = array([[1, 3], [5, 2], [0.1, 22]], "d")

        self.zeromtx = array(
            [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]], "d"
        )
        self.sparse1 = array(
            [[0.0, 0.0, 5.33], [0.0, 0.0, 0.4], [1.0, 0.0, 0.0], [0.0, 0.0, 0.0]], "d"
        )
        self.input_binary_dist_otu_gain1 = array(
            [[2, 1, 0, 0], [1, 0, 0, 1], [0, 0, 3, 0], [0, 0, 0, 1]]
        )

    def get_sym_mtx_from_uptri(self, mtx):
        """helper fn, only for square matrices"""
        numrows, numcols = shape(mtx)
        for i in range(numrows):
            for j in range(i):
                if i == j:
                    break
                mtx[i, j] = mtx[j, i]  # j < i, so row<col => upper triangle
        return mtx

    def test_dist_canberra(self):
        """tests dist_canberra

        tests inputs of empty mtx, zeros, and results compared with calcs done
        by hand"""

        assert_allclose(dist_canberra(self.zeromtx), zeros((4, 4), "d"))

        mtx1expected = array([[0.0, 46.2 / 52.2], [46.2 / 52.2, 0.0]], "d")
        assert_allclose(dist_canberra(self.mtx1), mtx1expected)

        sparse1exp = ones((self.sparse1.shape[0], self.sparse1.shape[0]))
        # remove diagonal
        sparse1exp[0, 0] = sparse1exp[1, 1] = sparse1exp[2, 2] = sparse1exp[3, 3] = 0.0

        sparse1exp[0, 1] = sparse1exp[1, 0] = (5.33 - 0.4) / (5.33 + 0.4)
        assert_allclose(dist_canberra(self.sparse1), sparse1exp)

    def test_dist_canberra_bug(self):
        i = array([[0, 0, 1], [0, 1, 1]])
        d = (1.0 / 2.0) * sum(
            [abs(0.0 - 1.0) / (0.0 + 1.0), abs(1.0 - 1.0) / (1.0 + 1.0)]
        )
        expected = array([[0.0, d], [d, 0.0]])
        actual = dist_canberra(i)
        assert_allclose(expected, actual)

    def test_dist_euclidean(self):
        """tests dist_euclidean

        tests inputs of empty mtx, zeros, and dense1 compared with calcs done
        by hand"""

        assert_allclose(dist_euclidean(self.zeromtx), zeros((4, 4), "d"))

        dense1expected = array(
            [
                [0.0, sqrt(17.0), sqrt(0.9**2 + 19**2)],
                [sqrt(17.0), 0.0, sqrt(4.9**2 + 20**2)],
                [sqrt(0.9**2 + 19**2), sqrt(4.9**2 + 20**2), 0.0],
            ],
            "d",
        )
        assert_allclose(dist_euclidean(self.dense1), dense1expected)

    def test_dist_gower(self):
        """tests dist_gower

        tests inputs of empty mtx, zeros, and results compared with calcs done
        by hand"""

        assert_allclose(dist_gower(self.zeromtx), zeros((4, 4), "d"))

        mtx1expected = array([[0.0, 2.0], [2.0, 0.0]], "d")
        assert_allclose(dist_gower(self.mtx1), mtx1expected)

        sparse1expected = array(
            [
                [0.0, 4.93 / 5.33, 2, 1],
                [4.93 / 5.33, 0.0, 1 + 0.4 / 5.33, 0.4 / 5.33],
                [2, 1 + 0.4 / 5.33, 0, 1],
                [1, 0.4 / 5.33, 1, 0.0],
            ],
            "d",
        )
        assert_allclose(dist_gower(self.sparse1), sparse1expected)

    def test_dist_manhattan(self):
        """tests dist_manhattan

        tests inputs of empty mtx, zeros, and dense1 compared with calcs done
        by hand"""

        assert_allclose(dist_manhattan(self.zeromtx), zeros((4, 4), "d"))

        dense1expected = array(
            [[0.0, 5.0, 019.9], [5.0, 0.0, 24.9], [19.9, 24.90, 0.0]], "d"
        )
        assert_allclose(dist_manhattan(self.dense1), dense1expected)

    def test_dist_abund_jaccard(self):
        """dist_abund_jaccard should compute distances for dense1 and mtx1"""
        mtx1_expected = array([[0, 0.25], [0.25, 0]], "d")
        assert_allclose(dist_abund_jaccard(self.mtx1), mtx1_expected)

        dense1_expected = zeros((3, 3), "d")
        assert_allclose(dist_abund_jaccard(self.dense1), dense1_expected)

        sparse1_expected = array(
            [
                [0.0, 0.0, 1.0, 1.0],
                [0.0, 0.0, 1.0, 1.0],
                [1.0, 1.0, 0.0, 1.0],
                [1.0, 1.0, 1.0, 0.0],
            ],
            "d",
        )
        assert_allclose(dist_abund_jaccard(self.sparse1), sparse1_expected)

    def test_dist_morisita_horn(self):
        """tests dist_morisita_horn

        tests inputs of empty mtx, zeros, and dense1 compared with calcs done
        by hand"""

        assert_allclose(dist_morisita_horn(self.zeromtx), zeros((4, 4), "d"))

        a = 1 - 2 * 69.3 / (26 / 16.0 * 23.1 * 4)
        mtx1expected = array([[0, a], [a, 0]], "d")
        assert_allclose(dist_morisita_horn(self.mtx1), mtx1expected)

    def test_dist_bray_curtis(self):
        """tests dist_bray_curtis

        tests inputs of empty mtx, zeros, and mtx1 compared with calcs done
        by hand"""

        assert_allclose(dist_manhattan(self.zeromtx), zeros((4, 4) * 1, "d"))

        mtx1expected = array([[0, 21.1 / 27.1], [21.1 / 27.1, 0]], "d")
        assert_allclose(dist_bray_curtis(self.mtx1), mtx1expected)

    def test_dist_bray_curtis_faith(self):
        """tests dist_bray_curtis_faith

        tests inputs of empty mtx, zeros, and mtx1 compared with calcs done
        by hand"""

        assert_allclose(dist_manhattan(self.zeromtx), zeros((4, 4) * 1, "d"))

        mtx1expected = array([[0, 21.1 / 27.1], [21.1 / 27.1, 0]], "d")
        assert_allclose(dist_bray_curtis_faith(self.mtx1), mtx1expected)

    def test_dist_soergel(self):
        """tests dist_soergel

        tests inputs of empty mtx, zeros, and dense1 compared with calcs done
        by hand/manhattan dist"""

        assert_allclose(dist_soergel(self.zeromtx), zeros((4, 4) * 1, "d"))

        dense1expected = dist_manhattan(self.dense1)
        dense1norm = array([[1, 8, 23], [8, 1, 27], [23, 27, 1]], "d")
        dense1expected /= dense1norm

        assert_allclose(dist_soergel(self.dense1), dense1expected)

    def test_dist_kulczynski(self):
        """tests dist_kulczynski

        tests inputs of empty mtx, zeros, and mtx1 compared with calcs done
        by hand"""

        assert_allclose(dist_kulczynski(self.zeromtx), zeros((4, 4) * 1, "d"))

        mtx1expected = array(
            [
                [0, 1.0 - 1.0 / 2.0 * (3.0 / 4.0 + 3.0 / 23.1)],
                [1.0 - 1.0 / 2.0 * (3.0 / 4.0 + 3.0 / 23.1), 0],
            ],
            "d",
        )

        assert_allclose(dist_kulczynski(self.mtx1), mtx1expected)

    def test_dist_pearson(self):
        """tests dist_pearson

        tests inputs of empty mtx, zeros, mtx compared with calcs done
        by hand, and an example from
        http://davidmlane.com/hyperstat/A56626.html
        """

        assert_allclose(dist_pearson(self.zeromtx), zeros((4, 4), "d"))

        mtx1expected = array([[0, 0], [0, 0]], "d")
        assert_allclose(dist_pearson(self.mtx1), mtx1expected)

        # example 1 from http://davidmlane.com/hyperstat/A56626.html
        ex1 = array([[1, 2, 3], [2, 5, 6]], "d")
        ex1res = 1 - 4.0 / sqrt(2.0 * (8 + 2.0 / 3.0))
        ex1expected = array([[0, ex1res], [ex1res, 0]], "d")

        assert_allclose(dist_pearson(ex1), ex1expected)

    def test_dist_spearman_approx(self):
        """tests dist_spearman_approx

        tests inputs of empty mtx, zeros, and an example from wikipedia
        """

        assert_allclose(dist_spearman_approx(self.zeromtx), zeros((4, 4) * 1, "d"))

        # ex1 from wikipedia Spearman's_rank_correlation_coefficient 20jan2009
        ex1 = array(
            [
                [106, 86, 100, 101, 99, 103, 97, 113, 112, 110],
                [7, 0, 27, 50, 28, 29, 20, 12, 6, 17],
            ],
            "d",
        )
        ex1res = 6.0 * 194.0 / (10.0 * 99.0)
        ex1expected = array([[0, ex1res], [ex1res, 0]], "d")
        assert_allclose(dist_spearman_approx(ex1), ex1expected)

    # now binary fns
    def test_binary_dist_otu_gain(self):
        """binary OTU gain functions as expected"""
        actual = binary_dist_otu_gain(self.input_binary_dist_otu_gain1)
        expected = array([[0, 1, 2, 2], [1, 0, 2, 1], [1, 1, 0, 1], [1, 0, 1, 0]])
        assert_allclose(actual, expected)

    def test_binary_dist_chisq(self):
        """tests binary_dist_chisq

        tests inputs of empty mtx, zeros, and mtx1 compared with calcs done
        by hand"""

        assert_allclose(binary_dist_chisq(self.zeromtx), zeros((4, 4), "d"))

        mtx1expected = array([[0, sqrt(9 / 8.0)], [sqrt(9 / 8.0), 0]], "d")
        assert_allclose(binary_dist_chisq(self.mtx1), mtx1expected)

    def test_binary_dist_chord(self):
        """tests binary_dist_chord

        tests inputs of empty mtx, zeros, and results compared with calcs done
        by hand"""

        assert_allclose(binary_dist_chord(self.zeromtx), zeros((4, 4), "d"))

        mtx1expected = array(
            [
                [0, sqrt(1 / 2.0 + (1.0 / sqrt(2.0) - 1.0) ** 2)],
                [sqrt(1 / 2.0 + (1.0 / sqrt(2.0) - 1.0) ** 2), 0],
            ],
            "d",
        )
        assert_allclose(binary_dist_chord(self.mtx1), mtx1expected)

    def test_binary_dist_lennon(self):
        """tests binary_dist_lennon

        tests inputs of empty mtx, zeros, and results compared with calcs done
        by hand"""

        assert_allclose(binary_dist_lennon(self.zeromtx), zeros((4, 4), "d"))

        mtxa = array([[5.2, 9, 0.2], [0, 99, 1], [0, 0.0, 8233.1]], "d")
        assert_allclose(binary_dist_lennon(mtxa), zeros((3, 3), "d"))

        mtxb = array([[5.2, 0, 0.2, 9.2], [0, 0, 0, 1], [0, 3.2, 0, 8233.1]], "d")
        mtxbexpected = array([[0, 0, 0.5], [0, 0, 0], [0.5, 0, 0]], "d")
        assert_allclose(binary_dist_lennon(mtxb), mtxbexpected)

    def test_binary_dist_pearson(self):
        """tests binary_dist_pearson

        tests inputs of empty mtx, zeros, and dense1 compared with calcs done
        by hand"""

        assert_allclose(binary_dist_pearson(self.zeromtx), zeros((4, 4), "d"))

        assert_allclose(binary_dist_pearson(self.dense1), zeros((3, 3)))

    def test_jaccard_set(self):
        """tests jaccard_set"""
        for klass in (set, frozenset):
            a = klass([1, 2, 3])
            b = klass([2, 3, 4])
            c = klass([4, 5, 6])
            empty = klass([])
            assert_allclose(jaccard(a, b), 2 / 4)
            assert_allclose(jaccard(a, a), 0)
            assert_allclose(jaccard(a, c), 1)
            assert_allclose(jaccard(a, empty), 1)
            assert_allclose(jaccard(empty, a), 1)
            assert_allclose(jaccard(empty, empty), 0)

        # compare 2 large (500 element) sets
        seed(0)
        values = [randint(0, 1_000_000) for _ in range(1000)]
        # choose 2 sets of 500 random items from the list
        a = set(sample(values, 500))
        b = set(sample(values, 500))
        distance = jaccard(a, b)
        assert_allclose(distance, 0.6715425)

    def test_jaccard_ndarray(self):
        """tests jaccard_ndarray"""
        a = array([1, 2, 3])
        b = array([2, 3, 4])
        c = array([4, 5, 6])
        empty = array([])
        assert_allclose(jaccard(a, b), 2 / 4)
        assert_allclose(jaccard(a, a), 0)
        assert_allclose(jaccard(a, c), 1)
        assert_allclose(jaccard(a, empty), 1)
        assert_allclose(jaccard(empty, a), 1)
        assert_allclose(jaccard(empty, empty), 0)

        # compare 2 large (500 element) sets
        seed(0)
        values = [randint(0, 1_000_000) for _ in range(1000)]
        # choose 2 sets of 500 random items from the list
        a = array([sample(values, 500)])
        b = array([sample(values, 500)])
        distance = jaccard(a, b)
        assert_allclose(distance, 0.6715425)

    def test_jaccard_other(self):
        """tests jaccard using types not covered with single dispatch"""
        for a, b in ((1, 2), ([1], [2]), ("a", "b"), (1.0, 2.0)):
            with self.assertRaises(NotImplementedError):
                jaccard(a, b)

    def test_binary_dist_ochiai(self):
        """tests binary_dist_ochiai

        tests inputs of empty mtx, zeros, and mtx1 compared with calcs done
        by hand"""

        assert_allclose(binary_dist_ochiai(self.zeromtx), zeros((4, 4), "d"))

        mtx1expected = array([[0, 1 - 1 / sqrt(2.0)], [1 - 1 / sqrt(2.0), 0]], "d")
        assert_allclose(binary_dist_ochiai(self.mtx1), mtx1expected)

    def test_binary_dist_hamming(self):
        """tests binary_dist_hamming

        tests inputs of empty mtx, zeros, and mtx1 compared with calcs done
        by hand"""

        assert_allclose(binary_dist_hamming(self.zeromtx), zeros((4, 4), "d"))

        mtx1expected = array([[0, 1], [1, 0]], "d")
        assert_allclose(binary_dist_hamming(self.mtx1), mtx1expected)

    def test_binary_dist_sorensen_dice(self):
        """tests binary_dist_sorensen_dice

        tests inputs of empty mtx, zeros, and mtx1 compared with calcs done
        by hand"""

        assert_allclose(binary_dist_sorensen_dice(self.zeromtx), zeros((4, 4), "d"))

        mtx1expected = array([[0, 1 / 3.0], [1 / 3.0, 0]], "d")
        assert_allclose(binary_dist_sorensen_dice(self.mtx1), mtx1expected)

        sparse1expected = array(
            [[0, 0, 1.0, 1.0], [0, 0, 1, 1], [1, 1, 0, 1], [1, 1, 1, 0]], "d"
        )

        assert_allclose(binary_dist_sorensen_dice(self.sparse1), sparse1expected)

    def test_binary_dist_euclidean(self):
        """tests binary_dist_euclidean

        tests two inputs compared with calculations by hand, and runs zeros
        and an empty input"""
        dense1expected = array([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]], "d")
        sparse1expected = zeros((4, 4), "d")
        sparse1expected[0, 2] = sqrt(2)
        sparse1expected[0, 3] = 1.0
        sparse1expected[1, 2] = sqrt(2)
        sparse1expected[1, 3] = 1.0
        sparse1expected[2, 3] = 1.0
        sparse1expected = self.get_sym_mtx_from_uptri(sparse1expected)

        assert_allclose(binary_dist_euclidean(self.dense1), dense1expected)
        assert_allclose(binary_dist_euclidean(self.sparse1), sparse1expected)
        assert_allclose(binary_dist_euclidean(self.zeromtx), zeros((4, 4), "d"))

    # zj's stuff
    def test_chord_transform(self):
        """trans_chord should return the exp result in the ref paper."""

        exp = [
            [0.40824829, 0.40824829, 0.81649658],
            [0.48507125, 0.72760688, 0.48507125],
            [0.90453403, 0.30151134, 0.30151134],
        ]
        res = trans_chord(self.mat_test)
        assert_allclose(res, exp)

    def test_chord_dist(self):
        """dist_chord should return the exp result."""

        assert_allclose(dist_chord(self.zeromtx), zeros((4, 4), "d"))

        exp = [
            [0.0, 0.46662021, 0.72311971],
            [0.46662021, 0.0, 0.62546036],
            [0.72311971, 0.62546036, 0.0],
        ]
        dist = dist_chord(self.mat_test)
        assert_allclose(dist, exp)

    def test_chisq_transform(self):
        """trans_chisq should return the exp result in the ref paper."""
        exp_m = [
            [0.42257713, 0.45643546, 0.84515425],
            [0.48294529, 0.7824608, 0.48294529],
            [1.01418511, 0.36514837, 0.3380617],
        ]
        res_m = trans_chisq(self.mat_test)
        assert_allclose(res_m, exp_m)

    def test_chisq_distance(self):
        """dist_chisq should return the exp result."""

        assert_allclose(dist_chisq(self.zeromtx), zeros((4, 4), "d"))

        exp_d = [
            [0.0, 0.4910521, 0.78452291],
            [0.4910521, 0.0, 0.69091002],
            [0.78452291, 0.69091002, 0.0],
        ]
        res_d = dist_chisq(self.mat_test)
        assert_allclose(res_d, exp_d)

    def test_hellinger_transform(self):
        """dist_hellinger should return the exp result in the ref paper."""
        exp = [
            [0.5, 0.5, 0.70710678],
            [0.53452248, 0.65465367, 0.53452248],
            [0.77459667, 0.4472136, 0.4472136],
        ]
        res = trans_hellinger(self.mat_test)
        assert_allclose(res, exp)

    def test_hellinger_distance(self):
        """dist_hellinger should return the exp result."""

        assert_allclose(dist_hellinger(self.zeromtx), zeros((4, 4), "d"))

        exp = [
            [0.0, 0.23429661, 0.38175149],
            [0.23429661, 0.0, 0.32907422],
            [0.38175149, 0.32907422, 0.0],
        ]
        dist = dist_hellinger(self.mat_test)
        assert_allclose(dist, exp)

    def test_species_profile_transform(self):
        """trans_specprof should return the exp result."""
        exp = [[0.25, 0.25, 0.5], [0.28571429, 0.42857143, 0.28571429], [0.6, 0.2, 0.2]]
        res = trans_specprof(self.mat_test)
        assert_allclose(res, exp)

    def test_species_profile_distance(self):
        """dist_specprof should return the exp result."""

        assert_allclose(dist_specprof(self.zeromtx), zeros((4, 4), "d"))

        exp = [
            [0.0, 0.28121457, 0.46368092],
            [0.28121457, 0.0, 0.39795395],
            [0.46368092, 0.39795395, 0.0],
        ]
        dist = dist_specprof(self.mat_test)
        assert_allclose(dist, exp)

    def test_dist_bray_curtis_magurran1(self):
        """zero values should return zero dist, or 1 with nonzero samples"""
        res = dist_bray_curtis_magurran(array([[0, 0, 0], [0, 0, 0], [1, 1, 1]]))
        assert_allclose(res, array([[0, 0, 1], [0, 0, 1], [1, 1, 0]]))

    def test_dist_bray_curtis_magurran2(self):
        """should match hand-calculated values"""
        res = dist_bray_curtis_magurran(array([[1, 4, 3], [1, 3, 5], [0, 2, 0]]))
        assert_allclose(
            res,
            array(
                [
                    [0, 1 - 14 / 17, 1 - (0.4)],
                    [1 - 14 / 17, 0, 1 - 4 / 11],
                    [1 - 0.4, 1 - 4 / 11, 0],
                ]
            ),
        )



import numba
import numpy

from numba import njit


# turn off code coverage as njit-ted code not accessible to coverage


@njit(numba.float64(numba.float64[:], numba.int64, numba.int64), cache=True)
def goertzel_inner(x, N, period):  # pragma: no cover
    """returns the power from series x for period"""
    coeff = 2.0 * numpy.cos(2 * numpy.pi / period)
    s_prev = 0.0
    s_prev2 = 0.0
    for n in range(N):
        s = x[n] + coeff * s_prev - s_prev2
        s_prev2 = s_prev
        s_prev = s

    power = numpy.sqrt(s_prev2**2 + s_prev**2 - coeff * s_prev2 * s_prev)
    return power


@njit(
    numba.complex128[:](
        numba.float64[:],
        numba.complex128[:],
        numba.complex128[:],
        numba.int64,
        numba.int64,
    ),
    cache=True,
)
def ipdft_inner(x, X, W, ulim, N):  # pragma: no cover
    for p in range(ulim):
        w = 1.0
        for n in range(N):
            if n != 0:
                w = w * W[p]
            X[p] = X[p] + x[n] * w
    return X


@njit(numba.void(numba.float64[:], numba.float64[:], numba.int64), cache=True)
def autocorr_inner(x, xc, N):  # pragma: no cover
    for m in range(-N + 1, N):
        for n in range(N):
            if 0 <= n - m < N:
                xc[m + N - 1] += x[n] * x[n - m]



"""
This file defines a class for controlling the scope and heterogeneity of
parameters involved in a maximum-likelihood based tree analysis.
"""

import pickle

import numpy

from cogent3.align import dp_calculation
from cogent3.align.pairwise import AlignableSeq
from cogent3.core.tree import TreeError
from cogent3.evolve import likelihood_calculation
from cogent3.evolve.likelihood_function import LikelihoodFunction as _LF
from cogent3.maths.stats.information_criteria import aic, bic
from cogent3.recalculation.scope import _indexed
from cogent3.util.misc import adjusted_gt_minprob


def _category_names(dimension, specified):
    if type(specified) is int:
        cats = [f"{dimension}{i}" for i in range(specified)]
    else:
        cats = tuple(specified)
    assert len(cats) >= 1, cats
    assert len(set(cats)) == len(cats), f"{dimension} names must be unique"
    return list(cats)


def load(filename):
    # first cut at saving pc's
    f = open(filename, "rb")
    (version, info, pc) = pickle.load(f)
    assert version < 2.0, version
    pc.update_intermediate_values()
    return pc


class _LikelihoodParameterController(_LF):
    """A ParameterController works by setting parameter rules. For each
    parameter in the model the edges of the tree are be partitioned into groups
    that share one value.

    For usage see the set_param_rule method.
    """

    # Basically wrapper around the more generic recalculation.ParameterController
    # class, which doesn't know about trees.

    def __init__(
        self,
        model,
        tree,
        bins=1,
        loci=1,
        optimise_motif_probs=False,
        motif_probs_from_align=False,
        default_length=1.0,
        **kw,
    ):
        # cache of arguments used to construct
        d = locals()
        exclude = ("self", "__class__", "kw")
        self._serialisable = {k: v for k, v in d.items() if k not in exclude}
        self._serialisable.update(kw)

        self.model = self._model = model
        self.tree = self._tree = tree
        self.seq_names = tree.get_tip_names()
        self.locus_names = _category_names("locus", loci)
        self.bin_names = _category_names("bin", bins)
        self.posn_names = [str(i) for i in range(model.word_length)]
        self.motifs = self._motifs = model.get_motifs()
        self._mprob_motifs = list(model.get_mprob_alphabet())
        defn = self.make_likelihood_defn(**kw)
        super(_LikelihoodParameterController, self).__init__(defn)
        self.set_default_param_rules()
        self.set_default_tree_parameter_rules(default_length=default_length)
        self.mprobs_from_alignment = motif_probs_from_align
        self.optimise_motif_probs = optimise_motif_probs
        self._name = None
        self._format = {}

    def save(self, filename):
        with open(filename, "w") as f:
            temp = {}
            try:
                for d in self.defns:
                    temp[id(d)] = d.values
                    del d.values
                pickle.dump((1.0, None, self), f)
            finally:
                for d in self.defns:
                    if id(d) in temp:
                        d.values = temp[id(d)]

    def set_default_tree_parameter_rules(self, default_length=1.0):
        """Lengths are set to the values found in the tree (if any), and
        free to be optimised independently.
        Other parameters are scoped based on the unique values found in the
        tree (if any) or default to having one value shared across the whole
        tree"""
        with self.updates_postponed():
            edges = self.tree.get_edge_vector()
            for par_name in self.model.get_param_list():
                try:
                    values = dict(
                        [
                            (edge.name, edge.params[par_name])
                            for edge in edges
                            if not edge.isroot()
                        ]
                    )
                    (uniq, index) = _indexed(values)
                except KeyError:
                    continue  # new parameter
                for u, value in enumerate(uniq):
                    group = [edge for (edge, i) in list(index.items()) if i == u]
                    self.set_param_rule(par_name, edges=group, init=value)

            # we are setting edge lengths, which is only valid if there is a
            # continuous-time model on the edge AND the edge is not the root
            if "expm" not in self.defn_for:  # discrete time only
                return

            exclude_edges = {"root"}
            if dp := self.defn_for.get("dpsubs", None):
                exclude_edges |= {n for _, n in dp.assignments.keys()}

            for edge in edges:
                if edge.name in exclude_edges:
                    continue
                length = edge.length or default_length
                self.set_param_rule("length", edge=edge.name, init=length)

    def set_motif_probs_from_data(
        self,
        align,
        locus=None,
        is_constant=None,
        include_ambiguity=False,
        is_independent=None,
        auto=False,
        warn=False,
        pseudocount=None,
        **kwargs,
    ):
        counts = self.model.count_motifs(align, include_ambiguity=include_ambiguity)
        if is_constant is None:
            is_constant = not self.optimise_motif_probs

        pseudocount = 0.0 if is_constant or (counts != 0).all() else pseudocount or 0.5
        counts += pseudocount
        mprobs = counts / (1.0 * sum(counts))
        self.set_motif_probs(
            mprobs,
            locus=locus,
            is_constant=is_constant,
            is_independent=is_independent,
            auto=auto,
            warn=warn,
            **kwargs,
        )

    def set_motif_probs(
        self,
        motif_probs,
        locus=None,
        bin=None,
        is_constant=None,
        is_independent=None,
        auto=False,
        warn=False,
        **kwargs,
    ):
        motif_probs = self.model.adapt_motif_probs(motif_probs, warn=warn)
        motif_probs = adjusted_gt_minprob(motif_probs, minprob=1e-6)
        if is_constant is None:
            is_constant = not self.optimise_motif_probs
        self.model.set_param_controller_motif_probs(
            self,
            motif_probs,
            is_constant=is_constant,
            bin=bin,
            locus=locus,
            is_independent=is_independent,
            **kwargs,
        )
        if not auto:
            self.mprobs_from_alignment = False  # should be done per-locus

    def set_expm(self, expm):
        assert expm in ["pade", "either", "eigen", "checked"], expm
        self.set_param_rule("expm", is_constant=True, value=expm)

    def make_calculator(self, **kw):
        return super(_LikelihoodParameterController, self).make_calculator(**kw)

    def _process_scope_info(
        self,
        edge=None,
        tip_names=None,
        edges=None,
        clade=None,
        stem=None,
        outgroup_name=None,
    ):
        """From information specifying the scope of a parameter derive a list of
        edge names"""

        if edges is not None:
            if tip_names or edge:
                raise TreeError("Only ONE of edge, edges or tip_names")
        elif edge is not None:
            if tip_names:
                raise TreeError("Only ONE of edge, edges or tip_names")
            edges = [edge]
        elif tip_names is None:
            edges = None  # meaning all edges
        elif len(tip_names) != 2:
            raise TreeError("tip_names must contain 2 species")
        else:
            (species1, species2) = tip_names
            if stem is None:
                stem = False
            if clade is None:
                clade = not stem
            edges = self.tree.get_edge_names(
                species1, species2, stem=stem, clade=clade, outgroup_name=outgroup_name
            )

        return edges

    def apply_param_rules(self, rules):
        """batch applies a collection of param rules"""
        with self.updates_postponed():
            for rule in rules:
                self.set_param_rule(**rule)

    def set_time_heterogeneity(
        self,
        exclude_params=None,
        edge_sets=None,
        is_independent=None,
        is_constant=False,
        value=None,
        lower=None,
        init=None,
        upper=None,
    ):
        """modifies the scope of substitution model rate params

        Parameters
        ----------
        exclude_params
            name(s) of substitution model predicate(s) to be excluded
        edge_sets
            series of dicts with an 'edges' key. Can also specify
            is_independent, is_constant etc.. If those are not provided, the
            method argument values are applied
        is_independent : bool
            whether edges in all edge sets are to be considered independent.
            default is False
            Overridden by edge_sets values.
        is_constant : bool
            makes constant all rate term parameters for all edge sets.
            Overridden by edge_sets values.
        value
            value for constant parameters, only valid when is_constant.
            Overridden by edge_sets values.
        lower, init, upper
            lower bound, starting value, upper bound for all parameters for
            all edge sets. Only valid if not is_constant.
            Overridden by edge_sets values.
        """
        if is_constant and any([lower, init, upper]):
            raise ValueError("cannot specify bounds or init for a constant param")

        if is_constant:
            kwargs = dict(is_constant=True, value=value)
        else:
            kwargs = dict(
                is_independent=is_independent, init=init, lower=lower, upper=upper
            )

        rate_terms = self._model.get_param_list()
        exclude_params = exclude_params or []
        if exclude_params and type(exclude_params) == str:
            exclude_params = [exclude_params]

        for param in exclude_params:
            if param not in rate_terms:
                raise ValueError(f"'{param}' not a valid rate param")

            rate_terms.remove(param)

        if edge_sets is None:
            # this just makes the following algorithm consistent
            # but we need to exclude any edges assigned to discrete psubs
            dpsubs = self.defn_for.get("dpsubs", None)
            exclude_edges = set()
            if dpsubs:
                dims = dpsubs.valid_dimensions
                index = dims.index("edge")
                for k in dpsubs.assignments:
                    exclude_edges.add(k[index])

            edge_sets = [
                dict(edges=[n])
                for n in self.tree.get_node_names(includeself=False)
                if n not in exclude_edges
            ]
        elif type(edge_sets) == dict:
            edge_sets = [edge_sets]

        # we make param rules
        param_rules = []
        for edge_set in edge_sets:
            edges = edge_set.get("edges", None)
            if type(edges) == str:
                edges = [edges]

            if edges:
                edges = list(edges)
                edge_set["edges"] = edges

            rule_base = kwargs.copy()
            rule_base.update(edge_set)
            for param in rate_terms:
                rule = rule_base.copy()
                rule.update(dict(par_name=param))
                param_rules.append(rule)

        self.apply_param_rules(param_rules)

    def set_param_rule(
        self,
        par_name,
        is_independent=None,
        is_constant=False,
        value=None,
        lower=None,
        init=None,
        upper=None,
        warn=False,
        **scope_info,
    ):
        """Define a model constraint for par_name. Parameters can be set
        constant or split according to tree/bin scopes.

        Parameters
        ----------
        par_name
            The model parameter being modified.
        is_constant, value
            if True, the parameter is held constant at
            value, if provided, or the likelihood functions current value.
        is_independent
            whether the partition specified by scope/bin
            arguments are to be considered independent.
        lower, init, upper
            specify the lower bound, initial value and
            upper bound for optimisation. Can be set separately.
        bin, bins
            the name(s) of the bin to apply rule.
        locus, loci
            the name of the locus/loci to apply rule.
        warn
            show warnings when numerical approximations used
        **scope_info
            tree scope arguments

              - edge, edges: The name of the tree edge(s) affected by rule.
              - tip_names: a tuple of two tip names, specifying a tree scope
                to apply rule.
              - outgroup_name: A tip name that, provided along with tip_names,
                ensures a consistently specified tree scope.
              - clade: The rule applies to all edges descending from the most
                recent common ancestor defined by the tip_names+outgroup_name
                arguments.
              - stem: The rule applies to the edge preceding the most recent
                common ancestor defined by the tip_names+outgroup_name
                arguments.
        """

        par_name = str(par_name)

        scopes = {}
        for single, plural in [
            ("bin", "bins"),
            ("locus", "loci"),
            ("position", "positions"),
            ("motif", "motifs"),
        ]:
            if single in scope_info:
                if v := scope_info.pop(single):
                    assert isinstance(v, str), f"{plural}=, maybe?"
                    assert plural not in scope_info
                    scopes[single] = [v]
            elif plural in scope_info:
                if v := scope_info.pop(plural):
                    scopes[single] = v

        if edges := self._process_scope_info(**scope_info):
            scopes["edge"] = edges

        if is_constant:
            assert not (init or lower or upper)
        elif init is not None:
            assert not value
            value = init
        self.assign_all(
            par_name,
            scopes,
            value,
            lower,
            upper,
            is_constant,
            is_independent,
            warn=warn,
        )

    def set_local_clock(self, tip1name, tip2name):
        """Constrain branch lengths for tip1name and tip2name to be equal.
        This is a molecular clock condition. Currently only valid for tips
        connected to the same node.

        Note: This is just a convenient interface to setParameterRule.
        """
        self.set_param_rule(
            "length", tip_names=[tip1name, tip2name], clade=True, is_independent=0
        )

    def set_constant_lengths(self, tree=None, exclude_list=None):
        """Constrains edge lengths to those in the tree.

        Parameters
        ----------
        tree
            must have the same topology as the current model.
            If not provided, the current tree length's are used.
        exclude_list
            a list of edge names whose branch lengths
            will be constrained.

        """

        exclude_list = exclude_list or []

        if tree is None:
            tree = self.tree

        with self.updates_postponed():
            for edge in tree.get_edge_vector():
                if edge.length is None or edge.name in exclude_list:
                    continue
                self.set_param_rule(
                    "length", edge=edge.name, is_constant=1, value=edge.length
                )

    def get_aic(self, second_order=False):
        """returns Aikake Information Criteria

        Parameters
        ----------
        second_order
            if true, the second
            adjusted by the alignment length

        """
        if second_order:
            sequence_length = sum(
                len(self.get_param_value("lht", locus=l).index)
                for l in self.locus_names
            )
        else:
            sequence_length = None

        lnL = self.get_log_likelihood()
        nfp = self.get_num_free_params()
        return aic(lnL, nfp, sequence_length)

    def get_bic(self):
        """returns the Bayesian Information Criteria"""
        sequence_length = sum(
            len(self.get_param_value("lht", locus=l).index) for l in self.locus_names
        )
        lnL = self.get_log_likelihood()
        nfp = self.get_num_free_params()
        return bic(lnL, nfp, sequence_length)


class AlignmentLikelihoodFunction(_LikelihoodParameterController):
    def set_default_param_rules(self):
        try:
            self.assign_all("fixed_motif", None, value=-1, const=True, independent=True)
        except KeyError:
            pass

    def make_likelihood_defn(self, sites_independent=True, discrete_edges=None):
        defns = self.model.make_param_controller_defns(bin_names=self.bin_names)
        if discrete_edges is not None:
            from .discrete_markov import PartialyDiscretePsubsDefn

            defns["psubs"] = PartialyDiscretePsubsDefn(
                self.motifs, defns["psubs"], discrete_edges
            )
        return likelihood_calculation.make_total_loglikelihood_defn(
            self.tree,
            defns["align"],
            defns["psubs"],
            defns["word_probs"],
            defns["bprobs"],
            self.bin_names,
            self.locus_names,
            sites_independent,
        )

    def set_alignment(self, aligns, motif_pseudocount=None):
        """set the alignment to be used for computing the likelihood."""
        if type(aligns) is not list:
            aligns = [aligns]
        assert len(aligns) == len(self.locus_names), len(aligns)
        tip_names = set(self.tree.get_tip_names())
        for index, aln in enumerate(aligns):
            if len(aligns) > 1:
                locus_name = f"for locus '{self.locus_names[index]}'"
            else:
                locus_name = ""
            assert not set(aln.names).symmetric_difference(
                tip_names
            ), "Tree tip names %s and aln seq names %s don't match %s" % (
                self.tree.get_tip_names(),
                aln.names,
                locus_name,
            )
            assert "root" not in aln.names, "'root' is a reserved name."
        with self.updates_postponed():
            for locus_name, align in zip(self.locus_names, aligns):
                self.assign_all(
                    "alignment", {"locus": [locus_name]}, value=align, const=True
                )
                if self.mprobs_from_alignment:
                    self.set_motif_probs_from_data(
                        align,
                        locus=locus_name,
                        auto=True,
                        pseudocount=motif_pseudocount,
                    )


class SequenceLikelihoodFunction(_LikelihoodParameterController):
    def set_default_param_rules(self):
        pass

    def make_likelihood_defn(
        self, sites_independent=None, with_indel_params=True, kn=True
    ):
        assert sites_independent is None or not sites_independent
        assert len(self.locus_names) == 1
        return dp_calculation.make_forward_tree_defn(
            self.model,
            self.tree,
            self.bin_names,
            with_indel_params=with_indel_params,
            kn=kn,
        )

    def set_sequences(self, seqs, locus=None):
        from cogent3.core.alignment import SequenceCollection

        leaves = {}

        if isinstance(seqs, SequenceCollection):
            seqs = seqs.named_seqs

        for name, seq in list(seqs.items()):
            # if has uniq, probably already a likelihood tree leaf obj already
            if hasattr(seq, "uniq"):
                # XXX more checks - same alphabet as model, name etc ...
                leaf = seq
            else:
                leaf = self.model.convert_sequence(seq, name)
            leaf = AlignableSeq(leaf)
            leaves[name] = leaf
            assert name != "root", "'root' is a reserved name."
        self.set_pogs(leaves, locus=locus)

    def set_pogs(self, leaves, locus=None):
        with self.updates_postponed():
            for name, pog in list(leaves.items()):
                self.set_param_rule("leaf", edge=name, value=pog, is_constant=True)
            if self.mprobs_from_alignment:
                counts = numpy.sum(
                    [pog.leaf.get_motif_counts() for pog in list(leaves.values())], 0
                )
                mprobs = counts / (1.0 * sum(counts))
                self.set_motif_probs(mprobs, locus=locus, is_constant=True, warn=False)
